/* autogenerated file */
#if !defined(SOCK_STREAM)
# if defined(HAVE_CONST_SOCK_STREAM)
#  define SOCK_STREAM SOCK_STREAM
# endif
#endif
#if !defined(SOCK_DGRAM)
# if defined(HAVE_CONST_SOCK_DGRAM)
#  define SOCK_DGRAM SOCK_DGRAM
# endif
#endif
#if !defined(SOCK_RAW)
# if defined(HAVE_CONST_SOCK_RAW)
#  define SOCK_RAW SOCK_RAW
# endif
#endif
#if !defined(SOCK_RDM)
# if defined(HAVE_CONST_SOCK_RDM)
#  define SOCK_RDM SOCK_RDM
# endif
#endif
#if !defined(SOCK_SEQPACKET)
# if defined(HAVE_CONST_SOCK_SEQPACKET)
#  define SOCK_SEQPACKET SOCK_SEQPACKET
# endif
#endif
#if !defined(SOCK_PACKET)
# if defined(HAVE_CONST_SOCK_PACKET)
#  define SOCK_PACKET SOCK_PACKET
# endif
#endif
#if !defined(SOCK_NONBLOCK)
# if defined(HAVE_CONST_SOCK_NONBLOCK)
#  define SOCK_NONBLOCK SOCK_NONBLOCK
# endif
#endif
#if !defined(SOCK_CLOEXEC)
# if defined(HAVE_CONST_SOCK_CLOEXEC)
#  define SOCK_CLOEXEC SOCK_CLOEXEC
# endif
#endif
#if !defined(AF_UNSPEC)
# if defined(HAVE_CONST_AF_UNSPEC)
#  define AF_UNSPEC AF_UNSPEC
# endif
#endif
#if !defined(PF_UNSPEC)
# if defined(HAVE_CONST_PF_UNSPEC)
#  define PF_UNSPEC PF_UNSPEC
# endif
#endif
#if !defined(AF_INET)
# if defined(HAVE_CONST_AF_INET)
#  define AF_INET AF_INET
# endif
#endif
#if !defined(PF_INET)
# if defined(HAVE_CONST_PF_INET)
#  define PF_INET PF_INET
# endif
#endif
#if !defined(AF_INET6)
# if defined(HAVE_CONST_AF_INET6)
#  define AF_INET6 AF_INET6
# endif
#endif
#if !defined(PF_INET6)
# if defined(HAVE_CONST_PF_INET6)
#  define PF_INET6 PF_INET6
# endif
#endif
#if !defined(AF_UNIX)
# if defined(HAVE_CONST_AF_UNIX)
#  define AF_UNIX AF_UNIX
# endif
#endif
#if !defined(PF_UNIX)
# if defined(HAVE_CONST_PF_UNIX)
#  define PF_UNIX PF_UNIX
# endif
#endif
#if !defined(AF_AX25)
# if defined(HAVE_CONST_AF_AX25)
#  define AF_AX25 AF_AX25
# endif
#endif
#if !defined(PF_AX25)
# if defined(HAVE_CONST_PF_AX25)
#  define PF_AX25 PF_AX25
# endif
#endif
#if !defined(AF_IPX)
# if defined(HAVE_CONST_AF_IPX)
#  define AF_IPX AF_IPX
# endif
#endif
#if !defined(PF_IPX)
# if defined(HAVE_CONST_PF_IPX)
#  define PF_IPX PF_IPX
# endif
#endif
#if !defined(AF_APPLETALK)
# if defined(HAVE_CONST_AF_APPLETALK)
#  define AF_APPLETALK AF_APPLETALK
# endif
#endif
#if !defined(PF_APPLETALK)
# if defined(HAVE_CONST_PF_APPLETALK)
#  define PF_APPLETALK PF_APPLETALK
# endif
#endif
#if !defined(AF_LOCAL)
# if defined(HAVE_CONST_AF_LOCAL)
#  define AF_LOCAL AF_LOCAL
# endif
#endif
#if !defined(PF_LOCAL)
# if defined(HAVE_CONST_PF_LOCAL)
#  define PF_LOCAL PF_LOCAL
# endif
#endif
#if !defined(AF_IMPLINK)
# if defined(HAVE_CONST_AF_IMPLINK)
#  define AF_IMPLINK AF_IMPLINK
# endif
#endif
#if !defined(PF_IMPLINK)
# if defined(HAVE_CONST_PF_IMPLINK)
#  define PF_IMPLINK PF_IMPLINK
# endif
#endif
#if !defined(AF_PUP)
# if defined(HAVE_CONST_AF_PUP)
#  define AF_PUP AF_PUP
# endif
#endif
#if !defined(PF_PUP)
# if defined(HAVE_CONST_PF_PUP)
#  define PF_PUP PF_PUP
# endif
#endif
#if !defined(AF_CHAOS)
# if defined(HAVE_CONST_AF_CHAOS)
#  define AF_CHAOS AF_CHAOS
# endif
#endif
#if !defined(PF_CHAOS)
# if defined(HAVE_CONST_PF_CHAOS)
#  define PF_CHAOS PF_CHAOS
# endif
#endif
#if !defined(AF_NS)
# if defined(HAVE_CONST_AF_NS)
#  define AF_NS AF_NS
# endif
#endif
#if !defined(PF_NS)
# if defined(HAVE_CONST_PF_NS)
#  define PF_NS PF_NS
# endif
#endif
#if !defined(AF_ISO)
# if defined(HAVE_CONST_AF_ISO)
#  define AF_ISO AF_ISO
# endif
#endif
#if !defined(PF_ISO)
# if defined(HAVE_CONST_PF_ISO)
#  define PF_ISO PF_ISO
# endif
#endif
#if !defined(AF_OSI)
# if defined(HAVE_CONST_AF_OSI)
#  define AF_OSI AF_OSI
# endif
#endif
#if !defined(PF_OSI)
# if defined(HAVE_CONST_PF_OSI)
#  define PF_OSI PF_OSI
# endif
#endif
#if !defined(AF_ECMA)
# if defined(HAVE_CONST_AF_ECMA)
#  define AF_ECMA AF_ECMA
# endif
#endif
#if !defined(PF_ECMA)
# if defined(HAVE_CONST_PF_ECMA)
#  define PF_ECMA PF_ECMA
# endif
#endif
#if !defined(AF_DATAKIT)
# if defined(HAVE_CONST_AF_DATAKIT)
#  define AF_DATAKIT AF_DATAKIT
# endif
#endif
#if !defined(PF_DATAKIT)
# if defined(HAVE_CONST_PF_DATAKIT)
#  define PF_DATAKIT PF_DATAKIT
# endif
#endif
#if !defined(AF_CCITT)
# if defined(HAVE_CONST_AF_CCITT)
#  define AF_CCITT AF_CCITT
# endif
#endif
#if !defined(PF_CCITT)
# if defined(HAVE_CONST_PF_CCITT)
#  define PF_CCITT PF_CCITT
# endif
#endif
#if !defined(AF_SNA)
# if defined(HAVE_CONST_AF_SNA)
#  define AF_SNA AF_SNA
# endif
#endif
#if !defined(PF_SNA)
# if defined(HAVE_CONST_PF_SNA)
#  define PF_SNA PF_SNA
# endif
#endif
#if !defined(AF_DEC)
# if defined(HAVE_CONST_AF_DEC)
#  define AF_DEC AF_DEC
# endif
#endif
#if !defined(PF_DEC)
# if defined(HAVE_CONST_PF_DEC)
#  define PF_DEC PF_DEC
# endif
#endif
#if !defined(AF_DECnet)
# if defined(HAVE_CONST_AF_DECNET)
#  define AF_DECnet AF_DECnet
# endif
#endif
#if !defined(PF_DECnet)
# if defined(HAVE_CONST_PF_DECNET)
#  define PF_DECnet PF_DECnet
# endif
#endif
#if !defined(AF_DLI)
# if defined(HAVE_CONST_AF_DLI)
#  define AF_DLI AF_DLI
# endif
#endif
#if !defined(PF_DLI)
# if defined(HAVE_CONST_PF_DLI)
#  define PF_DLI PF_DLI
# endif
#endif
#if !defined(AF_LAT)
# if defined(HAVE_CONST_AF_LAT)
#  define AF_LAT AF_LAT
# endif
#endif
#if !defined(PF_LAT)
# if defined(HAVE_CONST_PF_LAT)
#  define PF_LAT PF_LAT
# endif
#endif
#if !defined(AF_HYLINK)
# if defined(HAVE_CONST_AF_HYLINK)
#  define AF_HYLINK AF_HYLINK
# endif
#endif
#if !defined(PF_HYLINK)
# if defined(HAVE_CONST_PF_HYLINK)
#  define PF_HYLINK PF_HYLINK
# endif
#endif
#if !defined(AF_ROUTE)
# if defined(HAVE_CONST_AF_ROUTE)
#  define AF_ROUTE AF_ROUTE
# endif
#endif
#if !defined(PF_ROUTE)
# if defined(HAVE_CONST_PF_ROUTE)
#  define PF_ROUTE PF_ROUTE
# endif
#endif
#if !defined(AF_LINK)
# if defined(HAVE_CONST_AF_LINK)
#  define AF_LINK AF_LINK
# endif
#endif
#if !defined(PF_LINK)
# if defined(HAVE_CONST_PF_LINK)
#  define PF_LINK PF_LINK
# endif
#endif
#if !defined(AF_COIP)
# if defined(HAVE_CONST_AF_COIP)
#  define AF_COIP AF_COIP
# endif
#endif
#if !defined(PF_COIP)
# if defined(HAVE_CONST_PF_COIP)
#  define PF_COIP PF_COIP
# endif
#endif
#if !defined(AF_CNT)
# if defined(HAVE_CONST_AF_CNT)
#  define AF_CNT AF_CNT
# endif
#endif
#if !defined(PF_CNT)
# if defined(HAVE_CONST_PF_CNT)
#  define PF_CNT PF_CNT
# endif
#endif
#if !defined(AF_SIP)
# if defined(HAVE_CONST_AF_SIP)
#  define AF_SIP AF_SIP
# endif
#endif
#if !defined(PF_SIP)
# if defined(HAVE_CONST_PF_SIP)
#  define PF_SIP PF_SIP
# endif
#endif
#if !defined(AF_NDRV)
# if defined(HAVE_CONST_AF_NDRV)
#  define AF_NDRV AF_NDRV
# endif
#endif
#if !defined(PF_NDRV)
# if defined(HAVE_CONST_PF_NDRV)
#  define PF_NDRV PF_NDRV
# endif
#endif
#if !defined(AF_ISDN)
# if defined(HAVE_CONST_AF_ISDN)
#  define AF_ISDN AF_ISDN
# endif
#endif
#if !defined(PF_ISDN)
# if defined(HAVE_CONST_PF_ISDN)
#  define PF_ISDN PF_ISDN
# endif
#endif
#if !defined(AF_NATM)
# if defined(HAVE_CONST_AF_NATM)
#  define AF_NATM AF_NATM
# endif
#endif
#if !defined(PF_NATM)
# if defined(HAVE_CONST_PF_NATM)
#  define PF_NATM PF_NATM
# endif
#endif
#if !defined(AF_SYSTEM)
# if defined(HAVE_CONST_AF_SYSTEM)
#  define AF_SYSTEM AF_SYSTEM
# endif
#endif
#if !defined(PF_SYSTEM)
# if defined(HAVE_CONST_PF_SYSTEM)
#  define PF_SYSTEM PF_SYSTEM
# endif
#endif
#if !defined(AF_NETBIOS)
# if defined(HAVE_CONST_AF_NETBIOS)
#  define AF_NETBIOS AF_NETBIOS
# endif
#endif
#if !defined(PF_NETBIOS)
# if defined(HAVE_CONST_PF_NETBIOS)
#  define PF_NETBIOS PF_NETBIOS
# endif
#endif
#if !defined(AF_PPP)
# if defined(HAVE_CONST_AF_PPP)
#  define AF_PPP AF_PPP
# endif
#endif
#if !defined(PF_PPP)
# if defined(HAVE_CONST_PF_PPP)
#  define PF_PPP PF_PPP
# endif
#endif
#if !defined(AF_ATM)
# if defined(HAVE_CONST_AF_ATM)
#  define AF_ATM AF_ATM
# endif
#endif
#if !defined(PF_ATM)
# if defined(HAVE_CONST_PF_ATM)
#  define PF_ATM PF_ATM
# endif
#endif
#if !defined(AF_NETGRAPH)
# if defined(HAVE_CONST_AF_NETGRAPH)
#  define AF_NETGRAPH AF_NETGRAPH
# endif
#endif
#if !defined(PF_NETGRAPH)
# if defined(HAVE_CONST_PF_NETGRAPH)
#  define PF_NETGRAPH PF_NETGRAPH
# endif
#endif
#if !defined(AF_MAX)
# if defined(HAVE_CONST_AF_MAX)
#  define AF_MAX AF_MAX
# endif
#endif
#if !defined(PF_MAX)
# if defined(HAVE_CONST_PF_MAX)
#  define PF_MAX PF_MAX
# endif
#endif
#if !defined(AF_PACKET)
# if defined(HAVE_CONST_AF_PACKET)
#  define AF_PACKET AF_PACKET
# endif
#endif
#if !defined(PF_PACKET)
# if defined(HAVE_CONST_PF_PACKET)
#  define PF_PACKET PF_PACKET
# endif
#endif
#if !defined(AF_E164)
# if defined(HAVE_CONST_AF_E164)
#  define AF_E164 AF_E164
# endif
#endif
#if !defined(PF_XTP)
# if defined(HAVE_CONST_PF_XTP)
#  define PF_XTP PF_XTP
# endif
#endif
#if !defined(PF_RTIP)
# if defined(HAVE_CONST_PF_RTIP)
#  define PF_RTIP PF_RTIP
# endif
#endif
#if !defined(PF_PIP)
# if defined(HAVE_CONST_PF_PIP)
#  define PF_PIP PF_PIP
# endif
#endif
#if !defined(AF_KEY)
# if defined(HAVE_CONST_AF_KEY)
#  define AF_KEY AF_KEY
# endif
#endif
#if !defined(PF_KEY)
# if defined(HAVE_CONST_PF_KEY)
#  define PF_KEY PF_KEY
# endif
#endif
#if !defined(AF_NETLINK)
# if defined(HAVE_CONST_AF_NETLINK)
#  define AF_NETLINK AF_NETLINK
# endif
#endif
#if !defined(PF_NETLINK)
# if defined(HAVE_CONST_PF_NETLINK)
#  define PF_NETLINK PF_NETLINK
# endif
#endif
#if !defined(AF_RDS)
# if defined(HAVE_CONST_AF_RDS)
#  define AF_RDS AF_RDS
# endif
#endif
#if !defined(PF_RDS)
# if defined(HAVE_CONST_PF_RDS)
#  define PF_RDS PF_RDS
# endif
#endif
#if !defined(AF_PPPOX)
# if defined(HAVE_CONST_AF_PPPOX)
#  define AF_PPPOX AF_PPPOX
# endif
#endif
#if !defined(PF_PPPOX)
# if defined(HAVE_CONST_PF_PPPOX)
#  define PF_PPPOX PF_PPPOX
# endif
#endif
#if !defined(AF_LLC)
# if defined(HAVE_CONST_AF_LLC)
#  define AF_LLC AF_LLC
# endif
#endif
#if !defined(PF_LLC)
# if defined(HAVE_CONST_PF_LLC)
#  define PF_LLC PF_LLC
# endif
#endif
#if !defined(AF_IB)
# if defined(HAVE_CONST_AF_IB)
#  define AF_IB AF_IB
# endif
#endif
#if !defined(PF_IB)
# if defined(HAVE_CONST_PF_IB)
#  define PF_IB PF_IB
# endif
#endif
#if !defined(AF_MPLS)
# if defined(HAVE_CONST_AF_MPLS)
#  define AF_MPLS AF_MPLS
# endif
#endif
#if !defined(PF_MPLS)
# if defined(HAVE_CONST_PF_MPLS)
#  define PF_MPLS PF_MPLS
# endif
#endif
#if !defined(AF_CAN)
# if defined(HAVE_CONST_AF_CAN)
#  define AF_CAN AF_CAN
# endif
#endif
#if !defined(PF_CAN)
# if defined(HAVE_CONST_PF_CAN)
#  define PF_CAN PF_CAN
# endif
#endif
#if !defined(AF_TIPC)
# if defined(HAVE_CONST_AF_TIPC)
#  define AF_TIPC AF_TIPC
# endif
#endif
#if !defined(PF_TIPC)
# if defined(HAVE_CONST_PF_TIPC)
#  define PF_TIPC PF_TIPC
# endif
#endif
#if !defined(AF_BLUETOOTH)
# if defined(HAVE_CONST_AF_BLUETOOTH)
#  define AF_BLUETOOTH AF_BLUETOOTH
# endif
#endif
#if !defined(PF_BLUETOOTH)
# if defined(HAVE_CONST_PF_BLUETOOTH)
#  define PF_BLUETOOTH PF_BLUETOOTH
# endif
#endif
#if !defined(AF_ALG)
# if defined(HAVE_CONST_AF_ALG)
#  define AF_ALG AF_ALG
# endif
#endif
#if !defined(PF_ALG)
# if defined(HAVE_CONST_PF_ALG)
#  define PF_ALG PF_ALG
# endif
#endif
#if !defined(AF_VSOCK)
# if defined(HAVE_CONST_AF_VSOCK)
#  define AF_VSOCK AF_VSOCK
# endif
#endif
#if !defined(PF_VSOCK)
# if defined(HAVE_CONST_PF_VSOCK)
#  define PF_VSOCK PF_VSOCK
# endif
#endif
#if !defined(AF_KCM)
# if defined(HAVE_CONST_AF_KCM)
#  define AF_KCM AF_KCM
# endif
#endif
#if !defined(PF_KCM)
# if defined(HAVE_CONST_PF_KCM)
#  define PF_KCM PF_KCM
# endif
#endif
#if !defined(AF_XDP)
# if defined(HAVE_CONST_AF_XDP)
#  define AF_XDP AF_XDP
# endif
#endif
#if !defined(PF_XDP)
# if defined(HAVE_CONST_PF_XDP)
#  define PF_XDP PF_XDP
# endif
#endif
#if !defined(MSG_OOB)
# if defined(HAVE_CONST_MSG_OOB)
#  define MSG_OOB MSG_OOB
# endif
#endif
#if !defined(MSG_PEEK)
# if defined(HAVE_CONST_MSG_PEEK)
#  define MSG_PEEK MSG_PEEK
# endif
#endif
#if !defined(MSG_DONTROUTE)
# if defined(HAVE_CONST_MSG_DONTROUTE)
#  define MSG_DONTROUTE MSG_DONTROUTE
# endif
#endif
#if !defined(MSG_EOR)
# if defined(HAVE_CONST_MSG_EOR)
#  define MSG_EOR MSG_EOR
# endif
#endif
#if !defined(MSG_TRUNC)
# if defined(HAVE_CONST_MSG_TRUNC)
#  define MSG_TRUNC MSG_TRUNC
# endif
#endif
#if !defined(MSG_CTRUNC)
# if defined(HAVE_CONST_MSG_CTRUNC)
#  define MSG_CTRUNC MSG_CTRUNC
# endif
#endif
#if !defined(MSG_WAITALL)
# if defined(HAVE_CONST_MSG_WAITALL)
#  define MSG_WAITALL MSG_WAITALL
# endif
#endif
#if !defined(MSG_DONTWAIT)
# if defined(HAVE_CONST_MSG_DONTWAIT)
#  define MSG_DONTWAIT MSG_DONTWAIT
# endif
#endif
#if !defined(MSG_EOF)
# if defined(HAVE_CONST_MSG_EOF)
#  define MSG_EOF MSG_EOF
# endif
#endif
#if !defined(MSG_FLUSH)
# if defined(HAVE_CONST_MSG_FLUSH)
#  define MSG_FLUSH MSG_FLUSH
# endif
#endif
#if !defined(MSG_HOLD)
# if defined(HAVE_CONST_MSG_HOLD)
#  define MSG_HOLD MSG_HOLD
# endif
#endif
#if !defined(MSG_SEND)
# if defined(HAVE_CONST_MSG_SEND)
#  define MSG_SEND MSG_SEND
# endif
#endif
#if !defined(MSG_HAVEMORE)
# if defined(HAVE_CONST_MSG_HAVEMORE)
#  define MSG_HAVEMORE MSG_HAVEMORE
# endif
#endif
#if !defined(MSG_RCVMORE)
# if defined(HAVE_CONST_MSG_RCVMORE)
#  define MSG_RCVMORE MSG_RCVMORE
# endif
#endif
#if !defined(MSG_COMPAT)
# if defined(HAVE_CONST_MSG_COMPAT)
#  define MSG_COMPAT MSG_COMPAT
# endif
#endif
#if !defined(MSG_PROXY)
# if defined(HAVE_CONST_MSG_PROXY)
#  define MSG_PROXY MSG_PROXY
# endif
#endif
#if !defined(MSG_FIN)
# if defined(HAVE_CONST_MSG_FIN)
#  define MSG_FIN MSG_FIN
# endif
#endif
#if !defined(MSG_SYN)
# if defined(HAVE_CONST_MSG_SYN)
#  define MSG_SYN MSG_SYN
# endif
#endif
#if !defined(MSG_CONFIRM)
# if defined(HAVE_CONST_MSG_CONFIRM)
#  define MSG_CONFIRM MSG_CONFIRM
# endif
#endif
#if !defined(MSG_RST)
# if defined(HAVE_CONST_MSG_RST)
#  define MSG_RST MSG_RST
# endif
#endif
#if !defined(MSG_ERRQUEUE)
# if defined(HAVE_CONST_MSG_ERRQUEUE)
#  define MSG_ERRQUEUE MSG_ERRQUEUE
# endif
#endif
#if !defined(MSG_NOSIGNAL)
# if defined(HAVE_CONST_MSG_NOSIGNAL)
#  define MSG_NOSIGNAL MSG_NOSIGNAL
# endif
#endif
#if !defined(MSG_MORE)
# if defined(HAVE_CONST_MSG_MORE)
#  define MSG_MORE MSG_MORE
# endif
#endif
#if !defined(MSG_FASTOPEN)
# if defined(HAVE_CONST_MSG_FASTOPEN)
#  define MSG_FASTOPEN MSG_FASTOPEN
# endif
#endif
#if !defined(SOL_SOCKET)
# if defined(HAVE_CONST_SOL_SOCKET)
#  define SOL_SOCKET SOL_SOCKET
# endif
#endif
#if !defined(SOL_IP)
# if defined(HAVE_CONST_SOL_IP)
#  define SOL_IP SOL_IP
# endif
#endif
#if !defined(SOL_IPX)
# if defined(HAVE_CONST_SOL_IPX)
#  define SOL_IPX SOL_IPX
# endif
#endif
#if !defined(SOL_AX25)
# if defined(HAVE_CONST_SOL_AX25)
#  define SOL_AX25 SOL_AX25
# endif
#endif
#if !defined(SOL_ATALK)
# if defined(HAVE_CONST_SOL_ATALK)
#  define SOL_ATALK SOL_ATALK
# endif
#endif
#if !defined(SOL_TCP)
# if defined(HAVE_CONST_SOL_TCP)
#  define SOL_TCP SOL_TCP
# endif
#endif
#if !defined(SOL_UDP)
# if defined(HAVE_CONST_SOL_UDP)
#  define SOL_UDP SOL_UDP
# endif
#endif
#if !defined(IPPROTO_IP)
# if defined(HAVE_CONST_IPPROTO_IP)
#  define IPPROTO_IP IPPROTO_IP
# else
#  define IPPROTO_IP 0
# endif
#endif
#if !defined(IPPROTO_ICMP)
# if defined(HAVE_CONST_IPPROTO_ICMP)
#  define IPPROTO_ICMP IPPROTO_ICMP
# else
#  define IPPROTO_ICMP 1
# endif
#endif
#if !defined(IPPROTO_IGMP)
# if defined(HAVE_CONST_IPPROTO_IGMP)
#  define IPPROTO_IGMP IPPROTO_IGMP
# endif
#endif
#if !defined(IPPROTO_GGP)
# if defined(HAVE_CONST_IPPROTO_GGP)
#  define IPPROTO_GGP IPPROTO_GGP
# endif
#endif
#if !defined(IPPROTO_TCP)
# if defined(HAVE_CONST_IPPROTO_TCP)
#  define IPPROTO_TCP IPPROTO_TCP
# else
#  define IPPROTO_TCP 6
# endif
#endif
#if !defined(IPPROTO_EGP)
# if defined(HAVE_CONST_IPPROTO_EGP)
#  define IPPROTO_EGP IPPROTO_EGP
# endif
#endif
#if !defined(IPPROTO_PUP)
# if defined(HAVE_CONST_IPPROTO_PUP)
#  define IPPROTO_PUP IPPROTO_PUP
# endif
#endif
#if !defined(IPPROTO_UDP)
# if defined(HAVE_CONST_IPPROTO_UDP)
#  define IPPROTO_UDP IPPROTO_UDP
# else
#  define IPPROTO_UDP 17
# endif
#endif
#if !defined(IPPROTO_IDP)
# if defined(HAVE_CONST_IPPROTO_IDP)
#  define IPPROTO_IDP IPPROTO_IDP
# endif
#endif
#if !defined(IPPROTO_HELLO)
# if defined(HAVE_CONST_IPPROTO_HELLO)
#  define IPPROTO_HELLO IPPROTO_HELLO
# endif
#endif
#if !defined(IPPROTO_ND)
# if defined(HAVE_CONST_IPPROTO_ND)
#  define IPPROTO_ND IPPROTO_ND
# endif
#endif
#if !defined(IPPROTO_TP)
# if defined(HAVE_CONST_IPPROTO_TP)
#  define IPPROTO_TP IPPROTO_TP
# endif
#endif
#if !defined(IPPROTO_XTP)
# if defined(HAVE_CONST_IPPROTO_XTP)
#  define IPPROTO_XTP IPPROTO_XTP
# endif
#endif
#if !defined(IPPROTO_EON)
# if defined(HAVE_CONST_IPPROTO_EON)
#  define IPPROTO_EON IPPROTO_EON
# endif
#endif
#if !defined(IPPROTO_BIP)
# if defined(HAVE_CONST_IPPROTO_BIP)
#  define IPPROTO_BIP IPPROTO_BIP
# endif
#endif
#if !defined(IPPROTO_AH)
# if defined(HAVE_CONST_IPPROTO_AH)
#  define IPPROTO_AH IPPROTO_AH
# endif
#endif
#if !defined(IPPROTO_DSTOPTS)
# if defined(HAVE_CONST_IPPROTO_DSTOPTS)
#  define IPPROTO_DSTOPTS IPPROTO_DSTOPTS
# endif
#endif
#if !defined(IPPROTO_ESP)
# if defined(HAVE_CONST_IPPROTO_ESP)
#  define IPPROTO_ESP IPPROTO_ESP
# endif
#endif
#if !defined(IPPROTO_FRAGMENT)
# if defined(HAVE_CONST_IPPROTO_FRAGMENT)
#  define IPPROTO_FRAGMENT IPPROTO_FRAGMENT
# endif
#endif
#if !defined(IPPROTO_HOPOPTS)
# if defined(HAVE_CONST_IPPROTO_HOPOPTS)
#  define IPPROTO_HOPOPTS IPPROTO_HOPOPTS
# endif
#endif
#if !defined(IPPROTO_ICMPV6)
# if defined(HAVE_CONST_IPPROTO_ICMPV6)
#  define IPPROTO_ICMPV6 IPPROTO_ICMPV6
# endif
#endif
#if !defined(IPPROTO_IPV6)
# if defined(HAVE_CONST_IPPROTO_IPV6)
#  define IPPROTO_IPV6 IPPROTO_IPV6
# endif
#endif
#if !defined(IPPROTO_NONE)
# if defined(HAVE_CONST_IPPROTO_NONE)
#  define IPPROTO_NONE IPPROTO_NONE
# endif
#endif
#if !defined(IPPROTO_ROUTING)
# if defined(HAVE_CONST_IPPROTO_ROUTING)
#  define IPPROTO_ROUTING IPPROTO_ROUTING
# endif
#endif
#if !defined(IPPROTO_RAW)
# if defined(HAVE_CONST_IPPROTO_RAW)
#  define IPPROTO_RAW IPPROTO_RAW
# else
#  define IPPROTO_RAW 255
# endif
#endif
#if !defined(IPPROTO_MAX)
# if defined(HAVE_CONST_IPPROTO_MAX)
#  define IPPROTO_MAX IPPROTO_MAX
# endif
#endif
#if !defined(IPPORT_RESERVED)
# if defined(HAVE_CONST_IPPORT_RESERVED)
#  define IPPORT_RESERVED IPPORT_RESERVED
# else
#  define IPPORT_RESERVED 1024
# endif
#endif
#if !defined(IPPORT_USERRESERVED)
# if defined(HAVE_CONST_IPPORT_USERRESERVED)
#  define IPPORT_USERRESERVED IPPORT_USERRESERVED
# else
#  define IPPORT_USERRESERVED 5000
# endif
#endif
#if !defined(INADDR_ANY)
# if defined(HAVE_CONST_INADDR_ANY)
#  define INADDR_ANY INADDR_ANY
# else
#  define INADDR_ANY 0x00000000
# endif
#endif
#if !defined(INADDR_BROADCAST)
# if defined(HAVE_CONST_INADDR_BROADCAST)
#  define INADDR_BROADCAST INADDR_BROADCAST
# else
#  define INADDR_BROADCAST 0xffffffff
# endif
#endif
#if !defined(INADDR_LOOPBACK)
# if defined(HAVE_CONST_INADDR_LOOPBACK)
#  define INADDR_LOOPBACK INADDR_LOOPBACK
# else
#  define INADDR_LOOPBACK 0x7F000001
# endif
#endif
#if !defined(INADDR_UNSPEC_GROUP)
# if defined(HAVE_CONST_INADDR_UNSPEC_GROUP)
#  define INADDR_UNSPEC_GROUP INADDR_UNSPEC_GROUP
# else
#  define INADDR_UNSPEC_GROUP 0xe0000000
# endif
#endif
#if !defined(INADDR_ALLHOSTS_GROUP)
# if defined(HAVE_CONST_INADDR_ALLHOSTS_GROUP)
#  define INADDR_ALLHOSTS_GROUP INADDR_ALLHOSTS_GROUP
# else
#  define INADDR_ALLHOSTS_GROUP 0xe0000001
# endif
#endif
#if !defined(INADDR_MAX_LOCAL_GROUP)
# if defined(HAVE_CONST_INADDR_MAX_LOCAL_GROUP)
#  define INADDR_MAX_LOCAL_GROUP INADDR_MAX_LOCAL_GROUP
# else
#  define INADDR_MAX_LOCAL_GROUP 0xe00000ff
# endif
#endif
#if !defined(INADDR_NONE)
# if defined(HAVE_CONST_INADDR_NONE)
#  define INADDR_NONE INADDR_NONE
# else
#  define INADDR_NONE 0xffffffff
# endif
#endif
#if !defined(IP_OPTIONS)
# if defined(HAVE_CONST_IP_OPTIONS)
#  define IP_OPTIONS IP_OPTIONS
# endif
#endif
#if !defined(IP_HDRINCL)
# if defined(HAVE_CONST_IP_HDRINCL)
#  define IP_HDRINCL IP_HDRINCL
# endif
#endif
#if !defined(IP_TOS)
# if defined(HAVE_CONST_IP_TOS)
#  define IP_TOS IP_TOS
# endif
#endif
#if !defined(IP_TTL)
# if defined(HAVE_CONST_IP_TTL)
#  define IP_TTL IP_TTL
# endif
#endif
#if !defined(IP_RECVOPTS)
# if defined(HAVE_CONST_IP_RECVOPTS)
#  define IP_RECVOPTS IP_RECVOPTS
# endif
#endif
#if !defined(IP_RECVRETOPTS)
# if defined(HAVE_CONST_IP_RECVRETOPTS)
#  define IP_RECVRETOPTS IP_RECVRETOPTS
# endif
#endif
#if !defined(IP_RECVDSTADDR)
# if defined(HAVE_CONST_IP_RECVDSTADDR)
#  define IP_RECVDSTADDR IP_RECVDSTADDR
# endif
#endif
#if !defined(IP_RETOPTS)
# if defined(HAVE_CONST_IP_RETOPTS)
#  define IP_RETOPTS IP_RETOPTS
# endif
#endif
#if !defined(IP_MINTTL)
# if defined(HAVE_CONST_IP_MINTTL)
#  define IP_MINTTL IP_MINTTL
# endif
#endif
#if !defined(IP_DONTFRAG)
# if defined(HAVE_CONST_IP_DONTFRAG)
#  define IP_DONTFRAG IP_DONTFRAG
# endif
#endif
#if !defined(IP_SENDSRCADDR)
# if defined(HAVE_CONST_IP_SENDSRCADDR)
#  define IP_SENDSRCADDR IP_SENDSRCADDR
# endif
#endif
#if !defined(IP_ONESBCAST)
# if defined(HAVE_CONST_IP_ONESBCAST)
#  define IP_ONESBCAST IP_ONESBCAST
# endif
#endif
#if !defined(IP_RECVTTL)
# if defined(HAVE_CONST_IP_RECVTTL)
#  define IP_RECVTTL IP_RECVTTL
# endif
#endif
#if !defined(IP_RECVIF)
# if defined(HAVE_CONST_IP_RECVIF)
#  define IP_RECVIF IP_RECVIF
# endif
#endif
#if !defined(IP_RECVSLLA)
# if defined(HAVE_CONST_IP_RECVSLLA)
#  define IP_RECVSLLA IP_RECVSLLA
# endif
#endif
#if !defined(IP_PORTRANGE)
# if defined(HAVE_CONST_IP_PORTRANGE)
#  define IP_PORTRANGE IP_PORTRANGE
# endif
#endif
#if !defined(IP_MULTICAST_IF)
# if defined(HAVE_CONST_IP_MULTICAST_IF)
#  define IP_MULTICAST_IF IP_MULTICAST_IF
# endif
#endif
#if !defined(IP_MULTICAST_TTL)
# if defined(HAVE_CONST_IP_MULTICAST_TTL)
#  define IP_MULTICAST_TTL IP_MULTICAST_TTL
# endif
#endif
#if !defined(IP_MULTICAST_LOOP)
# if defined(HAVE_CONST_IP_MULTICAST_LOOP)
#  define IP_MULTICAST_LOOP IP_MULTICAST_LOOP
# endif
#endif
#if !defined(IP_ADD_MEMBERSHIP)
# if defined(HAVE_CONST_IP_ADD_MEMBERSHIP)
#  define IP_ADD_MEMBERSHIP IP_ADD_MEMBERSHIP
# endif
#endif
#if !defined(IP_DROP_MEMBERSHIP)
# if defined(HAVE_CONST_IP_DROP_MEMBERSHIP)
#  define IP_DROP_MEMBERSHIP IP_DROP_MEMBERSHIP
# endif
#endif
#if !defined(IP_DEFAULT_MULTICAST_TTL)
# if defined(HAVE_CONST_IP_DEFAULT_MULTICAST_TTL)
#  define IP_DEFAULT_MULTICAST_TTL IP_DEFAULT_MULTICAST_TTL
# endif
#endif
#if !defined(IP_DEFAULT_MULTICAST_LOOP)
# if defined(HAVE_CONST_IP_DEFAULT_MULTICAST_LOOP)
#  define IP_DEFAULT_MULTICAST_LOOP IP_DEFAULT_MULTICAST_LOOP
# endif
#endif
#if !defined(IP_MAX_MEMBERSHIPS)
# if defined(HAVE_CONST_IP_MAX_MEMBERSHIPS)
#  define IP_MAX_MEMBERSHIPS IP_MAX_MEMBERSHIPS
# endif
#endif
#if !defined(IP_ROUTER_ALERT)
# if defined(HAVE_CONST_IP_ROUTER_ALERT)
#  define IP_ROUTER_ALERT IP_ROUTER_ALERT
# endif
#endif
#if !defined(IP_PKTINFO)
# if defined(HAVE_CONST_IP_PKTINFO)
#  define IP_PKTINFO IP_PKTINFO
# endif
#endif
#if !defined(IP_PKTOPTIONS)
# if defined(HAVE_CONST_IP_PKTOPTIONS)
#  define IP_PKTOPTIONS IP_PKTOPTIONS
# endif
#endif
#if !defined(IP_MTU_DISCOVER)
# if defined(HAVE_CONST_IP_MTU_DISCOVER)
#  define IP_MTU_DISCOVER IP_MTU_DISCOVER
# endif
#endif
#if !defined(IP_RECVERR)
# if defined(HAVE_CONST_IP_RECVERR)
#  define IP_RECVERR IP_RECVERR
# endif
#endif
#if !defined(IP_RECVTOS)
# if defined(HAVE_CONST_IP_RECVTOS)
#  define IP_RECVTOS IP_RECVTOS
# endif
#endif
#if !defined(IP_MTU)
# if defined(HAVE_CONST_IP_MTU)
#  define IP_MTU IP_MTU
# endif
#endif
#if !defined(IP_FREEBIND)
# if defined(HAVE_CONST_IP_FREEBIND)
#  define IP_FREEBIND IP_FREEBIND
# endif
#endif
#if !defined(IP_IPSEC_POLICY)
# if defined(HAVE_CONST_IP_IPSEC_POLICY)
#  define IP_IPSEC_POLICY IP_IPSEC_POLICY
# endif
#endif
#if !defined(IP_XFRM_POLICY)
# if defined(HAVE_CONST_IP_XFRM_POLICY)
#  define IP_XFRM_POLICY IP_XFRM_POLICY
# endif
#endif
#if !defined(IP_PASSSEC)
# if defined(HAVE_CONST_IP_PASSSEC)
#  define IP_PASSSEC IP_PASSSEC
# endif
#endif
#if !defined(IP_TRANSPARENT)
# if defined(HAVE_CONST_IP_TRANSPARENT)
#  define IP_TRANSPARENT IP_TRANSPARENT
# endif
#endif
#if !defined(IP_PMTUDISC_DONT)
# if defined(HAVE_CONST_IP_PMTUDISC_DONT)
#  define IP_PMTUDISC_DONT IP_PMTUDISC_DONT
# endif
#endif
#if !defined(IP_PMTUDISC_WANT)
# if defined(HAVE_CONST_IP_PMTUDISC_WANT)
#  define IP_PMTUDISC_WANT IP_PMTUDISC_WANT
# endif
#endif
#if !defined(IP_PMTUDISC_DO)
# if defined(HAVE_CONST_IP_PMTUDISC_DO)
#  define IP_PMTUDISC_DO IP_PMTUDISC_DO
# endif
#endif
#if !defined(IP_UNBLOCK_SOURCE)
# if defined(HAVE_CONST_IP_UNBLOCK_SOURCE)
#  define IP_UNBLOCK_SOURCE IP_UNBLOCK_SOURCE
# endif
#endif
#if !defined(IP_BLOCK_SOURCE)
# if defined(HAVE_CONST_IP_BLOCK_SOURCE)
#  define IP_BLOCK_SOURCE IP_BLOCK_SOURCE
# endif
#endif
#if !defined(IP_ADD_SOURCE_MEMBERSHIP)
# if defined(HAVE_CONST_IP_ADD_SOURCE_MEMBERSHIP)
#  define IP_ADD_SOURCE_MEMBERSHIP IP_ADD_SOURCE_MEMBERSHIP
# endif
#endif
#if !defined(IP_DROP_SOURCE_MEMBERSHIP)
# if defined(HAVE_CONST_IP_DROP_SOURCE_MEMBERSHIP)
#  define IP_DROP_SOURCE_MEMBERSHIP IP_DROP_SOURCE_MEMBERSHIP
# endif
#endif
#if !defined(IP_MSFILTER)
# if defined(HAVE_CONST_IP_MSFILTER)
#  define IP_MSFILTER IP_MSFILTER
# endif
#endif
#if !defined(MCAST_JOIN_GROUP)
# if defined(HAVE_CONST_MCAST_JOIN_GROUP)
#  define MCAST_JOIN_GROUP MCAST_JOIN_GROUP
# endif
#endif
#if !defined(MCAST_BLOCK_SOURCE)
# if defined(HAVE_CONST_MCAST_BLOCK_SOURCE)
#  define MCAST_BLOCK_SOURCE MCAST_BLOCK_SOURCE
# endif
#endif
#if !defined(MCAST_UNBLOCK_SOURCE)
# if defined(HAVE_CONST_MCAST_UNBLOCK_SOURCE)
#  define MCAST_UNBLOCK_SOURCE MCAST_UNBLOCK_SOURCE
# endif
#endif
#if !defined(MCAST_LEAVE_GROUP)
# if defined(HAVE_CONST_MCAST_LEAVE_GROUP)
#  define MCAST_LEAVE_GROUP MCAST_LEAVE_GROUP
# endif
#endif
#if !defined(MCAST_JOIN_SOURCE_GROUP)
# if defined(HAVE_CONST_MCAST_JOIN_SOURCE_GROUP)
#  define MCAST_JOIN_SOURCE_GROUP MCAST_JOIN_SOURCE_GROUP
# endif
#endif
#if !defined(MCAST_LEAVE_SOURCE_GROUP)
# if defined(HAVE_CONST_MCAST_LEAVE_SOURCE_GROUP)
#  define MCAST_LEAVE_SOURCE_GROUP MCAST_LEAVE_SOURCE_GROUP
# endif
#endif
#if !defined(MCAST_MSFILTER)
# if defined(HAVE_CONST_MCAST_MSFILTER)
#  define MCAST_MSFILTER MCAST_MSFILTER
# endif
#endif
#if !defined(MCAST_EXCLUDE)
# if defined(HAVE_CONST_MCAST_EXCLUDE)
#  define MCAST_EXCLUDE MCAST_EXCLUDE
# endif
#endif
#if !defined(MCAST_INCLUDE)
# if defined(HAVE_CONST_MCAST_INCLUDE)
#  define MCAST_INCLUDE MCAST_INCLUDE
# endif
#endif
#if !defined(SO_DEBUG)
# if defined(HAVE_CONST_SO_DEBUG)
#  define SO_DEBUG SO_DEBUG
# endif
#endif
#if !defined(SO_REUSEADDR)
# if defined(HAVE_CONST_SO_REUSEADDR)
#  define SO_REUSEADDR SO_REUSEADDR
# endif
#endif
#if !defined(SO_REUSEPORT)
# if defined(HAVE_CONST_SO_REUSEPORT)
#  define SO_REUSEPORT SO_REUSEPORT
# endif
#endif
#if !defined(SO_TYPE)
# if defined(HAVE_CONST_SO_TYPE)
#  define SO_TYPE SO_TYPE
# endif
#endif
#if !defined(SO_ERROR)
# if defined(HAVE_CONST_SO_ERROR)
#  define SO_ERROR SO_ERROR
# endif
#endif
#if !defined(SO_DONTROUTE)
# if defined(HAVE_CONST_SO_DONTROUTE)
#  define SO_DONTROUTE SO_DONTROUTE
# endif
#endif
#if !defined(SO_BROADCAST)
# if defined(HAVE_CONST_SO_BROADCAST)
#  define SO_BROADCAST SO_BROADCAST
# endif
#endif
#if !defined(SO_SNDBUF)
# if defined(HAVE_CONST_SO_SNDBUF)
#  define SO_SNDBUF SO_SNDBUF
# endif
#endif
#if !defined(SO_RCVBUF)
# if defined(HAVE_CONST_SO_RCVBUF)
#  define SO_RCVBUF SO_RCVBUF
# endif
#endif
#if !defined(SO_SNDBUFFORCE)
# if defined(HAVE_CONST_SO_SNDBUFFORCE)
#  define SO_SNDBUFFORCE SO_SNDBUFFORCE
# endif
#endif
#if !defined(SO_RCVBUFFORCE)
# if defined(HAVE_CONST_SO_RCVBUFFORCE)
#  define SO_RCVBUFFORCE SO_RCVBUFFORCE
# endif
#endif
#if !defined(SO_KEEPALIVE)
# if defined(HAVE_CONST_SO_KEEPALIVE)
#  define SO_KEEPALIVE SO_KEEPALIVE
# endif
#endif
#if !defined(SO_OOBINLINE)
# if defined(HAVE_CONST_SO_OOBINLINE)
#  define SO_OOBINLINE SO_OOBINLINE
# endif
#endif
#if !defined(SO_NO_CHECK)
# if defined(HAVE_CONST_SO_NO_CHECK)
#  define SO_NO_CHECK SO_NO_CHECK
# endif
#endif
#if !defined(SO_PRIORITY)
# if defined(HAVE_CONST_SO_PRIORITY)
#  define SO_PRIORITY SO_PRIORITY
# endif
#endif
#if !defined(SO_LINGER)
# if defined(HAVE_CONST_SO_LINGER)
#  define SO_LINGER SO_LINGER
# endif
#endif
#if !defined(SO_PASSCRED)
# if defined(HAVE_CONST_SO_PASSCRED)
#  define SO_PASSCRED SO_PASSCRED
# endif
#endif
#if !defined(SO_PEERCRED)
# if defined(HAVE_CONST_SO_PEERCRED)
#  define SO_PEERCRED SO_PEERCRED
# endif
#endif
#if !defined(SO_RCVLOWAT)
# if defined(HAVE_CONST_SO_RCVLOWAT)
#  define SO_RCVLOWAT SO_RCVLOWAT
# endif
#endif
#if !defined(SO_SNDLOWAT)
# if defined(HAVE_CONST_SO_SNDLOWAT)
#  define SO_SNDLOWAT SO_SNDLOWAT
# endif
#endif
#if !defined(SO_RCVTIMEO)
# if defined(HAVE_CONST_SO_RCVTIMEO)
#  define SO_RCVTIMEO SO_RCVTIMEO
# endif
#endif
#if !defined(SO_SNDTIMEO)
# if defined(HAVE_CONST_SO_SNDTIMEO)
#  define SO_SNDTIMEO SO_SNDTIMEO
# endif
#endif
#if !defined(SO_ACCEPTCONN)
# if defined(HAVE_CONST_SO_ACCEPTCONN)
#  define SO_ACCEPTCONN SO_ACCEPTCONN
# endif
#endif
#if !defined(SO_USELOOPBACK)
# if defined(HAVE_CONST_SO_USELOOPBACK)
#  define SO_USELOOPBACK SO_USELOOPBACK
# endif
#endif
#if !defined(SO_ACCEPTFILTER)
# if defined(HAVE_CONST_SO_ACCEPTFILTER)
#  define SO_ACCEPTFILTER SO_ACCEPTFILTER
# endif
#endif
#if !defined(SO_DONTTRUNC)
# if defined(HAVE_CONST_SO_DONTTRUNC)
#  define SO_DONTTRUNC SO_DONTTRUNC
# endif
#endif
#if !defined(SO_WANTMORE)
# if defined(HAVE_CONST_SO_WANTMORE)
#  define SO_WANTMORE SO_WANTMORE
# endif
#endif
#if !defined(SO_WANTOOBFLAG)
# if defined(HAVE_CONST_SO_WANTOOBFLAG)
#  define SO_WANTOOBFLAG SO_WANTOOBFLAG
# endif
#endif
#if !defined(SO_NREAD)
# if defined(HAVE_CONST_SO_NREAD)
#  define SO_NREAD SO_NREAD
# endif
#endif
#if !defined(SO_NKE)
# if defined(HAVE_CONST_SO_NKE)
#  define SO_NKE SO_NKE
# endif
#endif
#if !defined(SO_NOSIGPIPE)
# if defined(HAVE_CONST_SO_NOSIGPIPE)
#  define SO_NOSIGPIPE SO_NOSIGPIPE
# endif
#endif
#if !defined(SO_SECURITY_AUTHENTICATION)
# if defined(HAVE_CONST_SO_SECURITY_AUTHENTICATION)
#  define SO_SECURITY_AUTHENTICATION SO_SECURITY_AUTHENTICATION
# endif
#endif
#if !defined(SO_SECURITY_ENCRYPTION_TRANSPORT)
# if defined(HAVE_CONST_SO_SECURITY_ENCRYPTION_TRANSPORT)
#  define SO_SECURITY_ENCRYPTION_TRANSPORT SO_SECURITY_ENCRYPTION_TRANSPORT
# endif
#endif
#if !defined(SO_SECURITY_ENCRYPTION_NETWORK)
# if defined(HAVE_CONST_SO_SECURITY_ENCRYPTION_NETWORK)
#  define SO_SECURITY_ENCRYPTION_NETWORK SO_SECURITY_ENCRYPTION_NETWORK
# endif
#endif
#if !defined(SO_BINDTODEVICE)
# if defined(HAVE_CONST_SO_BINDTODEVICE)
#  define SO_BINDTODEVICE SO_BINDTODEVICE
# endif
#endif
#if !defined(SO_ATTACH_FILTER)
# if defined(HAVE_CONST_SO_ATTACH_FILTER)
#  define SO_ATTACH_FILTER SO_ATTACH_FILTER
# endif
#endif
#if !defined(SO_DETACH_FILTER)
# if defined(HAVE_CONST_SO_DETACH_FILTER)
#  define SO_DETACH_FILTER SO_DETACH_FILTER
# endif
#endif
#if !defined(SO_GET_FILTER)
# if defined(HAVE_CONST_SO_GET_FILTER)
#  define SO_GET_FILTER SO_GET_FILTER
# endif
#endif
#if !defined(SO_PEERNAME)
# if defined(HAVE_CONST_SO_PEERNAME)
#  define SO_PEERNAME SO_PEERNAME
# endif
#endif
#if !defined(SO_TIMESTAMP)
# if defined(HAVE_CONST_SO_TIMESTAMP)
#  define SO_TIMESTAMP SO_TIMESTAMP
# endif
#endif
#if !defined(SO_TIMESTAMPNS)
# if defined(HAVE_CONST_SO_TIMESTAMPNS)
#  define SO_TIMESTAMPNS SO_TIMESTAMPNS
# endif
#endif
#if !defined(SO_BINTIME)
# if defined(HAVE_CONST_SO_BINTIME)
#  define SO_BINTIME SO_BINTIME
# endif
#endif
#if !defined(SO_RECVUCRED)
# if defined(HAVE_CONST_SO_RECVUCRED)
#  define SO_RECVUCRED SO_RECVUCRED
# endif
#endif
#if !defined(SO_MAC_EXEMPT)
# if defined(HAVE_CONST_SO_MAC_EXEMPT)
#  define SO_MAC_EXEMPT SO_MAC_EXEMPT
# endif
#endif
#if !defined(SO_ALLZONES)
# if defined(HAVE_CONST_SO_ALLZONES)
#  define SO_ALLZONES SO_ALLZONES
# endif
#endif
#if !defined(SO_PEERSEC)
# if defined(HAVE_CONST_SO_PEERSEC)
#  define SO_PEERSEC SO_PEERSEC
# endif
#endif
#if !defined(SO_PASSSEC)
# if defined(HAVE_CONST_SO_PASSSEC)
#  define SO_PASSSEC SO_PASSSEC
# endif
#endif
#if !defined(SO_MARK)
# if defined(HAVE_CONST_SO_MARK)
#  define SO_MARK SO_MARK
# endif
#endif
#if !defined(SO_TIMESTAMPING)
# if defined(HAVE_CONST_SO_TIMESTAMPING)
#  define SO_TIMESTAMPING SO_TIMESTAMPING
# endif
#endif
#if !defined(SO_PROTOCOL)
# if defined(HAVE_CONST_SO_PROTOCOL)
#  define SO_PROTOCOL SO_PROTOCOL
# endif
#endif
#if !defined(SO_DOMAIN)
# if defined(HAVE_CONST_SO_DOMAIN)
#  define SO_DOMAIN SO_DOMAIN
# endif
#endif
#if !defined(SO_RXQ_OVFL)
# if defined(HAVE_CONST_SO_RXQ_OVFL)
#  define SO_RXQ_OVFL SO_RXQ_OVFL
# endif
#endif
#if !defined(SO_WIFI_STATUS)
# if defined(HAVE_CONST_SO_WIFI_STATUS)
#  define SO_WIFI_STATUS SO_WIFI_STATUS
# endif
#endif
#if !defined(SO_PEEK_OFF)
# if defined(HAVE_CONST_SO_PEEK_OFF)
#  define SO_PEEK_OFF SO_PEEK_OFF
# endif
#endif
#if !defined(SO_NOFCS)
# if defined(HAVE_CONST_SO_NOFCS)
#  define SO_NOFCS SO_NOFCS
# endif
#endif
#if !defined(SO_LOCK_FILTER)
# if defined(HAVE_CONST_SO_LOCK_FILTER)
#  define SO_LOCK_FILTER SO_LOCK_FILTER
# endif
#endif
#if !defined(SO_SELECT_ERR_QUEUE)
# if defined(HAVE_CONST_SO_SELECT_ERR_QUEUE)
#  define SO_SELECT_ERR_QUEUE SO_SELECT_ERR_QUEUE
# endif
#endif
#if !defined(SO_BUSY_POLL)
# if defined(HAVE_CONST_SO_BUSY_POLL)
#  define SO_BUSY_POLL SO_BUSY_POLL
# endif
#endif
#if !defined(SO_MAX_PACING_RATE)
# if defined(HAVE_CONST_SO_MAX_PACING_RATE)
#  define SO_MAX_PACING_RATE SO_MAX_PACING_RATE
# endif
#endif
#if !defined(SO_BPF_EXTENSIONS)
# if defined(HAVE_CONST_SO_BPF_EXTENSIONS)
#  define SO_BPF_EXTENSIONS SO_BPF_EXTENSIONS
# endif
#endif
#if !defined(SOPRI_INTERACTIVE)
# if defined(HAVE_CONST_SOPRI_INTERACTIVE)
#  define SOPRI_INTERACTIVE SOPRI_INTERACTIVE
# endif
#endif
#if !defined(SOPRI_NORMAL)
# if defined(HAVE_CONST_SOPRI_NORMAL)
#  define SOPRI_NORMAL SOPRI_NORMAL
# endif
#endif
#if !defined(SOPRI_BACKGROUND)
# if defined(HAVE_CONST_SOPRI_BACKGROUND)
#  define SOPRI_BACKGROUND SOPRI_BACKGROUND
# endif
#endif
#if !defined(IPX_TYPE)
# if defined(HAVE_CONST_IPX_TYPE)
#  define IPX_TYPE IPX_TYPE
# endif
#endif
#if !defined(TCP_NODELAY)
# if defined(HAVE_CONST_TCP_NODELAY)
#  define TCP_NODELAY TCP_NODELAY
# endif
#endif
#if !defined(TCP_MAXSEG)
# if defined(HAVE_CONST_TCP_MAXSEG)
#  define TCP_MAXSEG TCP_MAXSEG
# endif
#endif
#if !defined(TCP_CORK)
# if defined(HAVE_CONST_TCP_CORK)
#  define TCP_CORK TCP_CORK
# endif
#endif
#if !defined(TCP_DEFER_ACCEPT)
# if defined(HAVE_CONST_TCP_DEFER_ACCEPT)
#  define TCP_DEFER_ACCEPT TCP_DEFER_ACCEPT
# endif
#endif
#if !defined(TCP_INFO)
# if defined(HAVE_CONST_TCP_INFO)
#  define TCP_INFO TCP_INFO
# endif
#endif
#if !defined(TCP_KEEPCNT)
# if defined(HAVE_CONST_TCP_KEEPCNT)
#  define TCP_KEEPCNT TCP_KEEPCNT
# endif
#endif
#if !defined(TCP_KEEPIDLE)
# if defined(HAVE_CONST_TCP_KEEPIDLE)
#  define TCP_KEEPIDLE TCP_KEEPIDLE
# endif
#endif
#if !defined(TCP_KEEPINTVL)
# if defined(HAVE_CONST_TCP_KEEPINTVL)
#  define TCP_KEEPINTVL TCP_KEEPINTVL
# endif
#endif
#if !defined(TCP_LINGER2)
# if defined(HAVE_CONST_TCP_LINGER2)
#  define TCP_LINGER2 TCP_LINGER2
# endif
#endif
#if !defined(TCP_MD5SIG)
# if defined(HAVE_CONST_TCP_MD5SIG)
#  define TCP_MD5SIG TCP_MD5SIG
# endif
#endif
#if !defined(TCP_NOOPT)
# if defined(HAVE_CONST_TCP_NOOPT)
#  define TCP_NOOPT TCP_NOOPT
# endif
#endif
#if !defined(TCP_NOPUSH)
# if defined(HAVE_CONST_TCP_NOPUSH)
#  define TCP_NOPUSH TCP_NOPUSH
# endif
#endif
#if !defined(TCP_QUICKACK)
# if defined(HAVE_CONST_TCP_QUICKACK)
#  define TCP_QUICKACK TCP_QUICKACK
# endif
#endif
#if !defined(TCP_SYNCNT)
# if defined(HAVE_CONST_TCP_SYNCNT)
#  define TCP_SYNCNT TCP_SYNCNT
# endif
#endif
#if !defined(TCP_WINDOW_CLAMP)
# if defined(HAVE_CONST_TCP_WINDOW_CLAMP)
#  define TCP_WINDOW_CLAMP TCP_WINDOW_CLAMP
# endif
#endif
#if !defined(TCP_FASTOPEN)
# if defined(HAVE_CONST_TCP_FASTOPEN)
#  define TCP_FASTOPEN TCP_FASTOPEN
# endif
#endif
#if !defined(TCP_CONGESTION)
# if defined(HAVE_CONST_TCP_CONGESTION)
#  define TCP_CONGESTION TCP_CONGESTION
# endif
#endif
#if !defined(TCP_COOKIE_TRANSACTIONS)
# if defined(HAVE_CONST_TCP_COOKIE_TRANSACTIONS)
#  define TCP_COOKIE_TRANSACTIONS TCP_COOKIE_TRANSACTIONS
# endif
#endif
#if !defined(TCP_QUEUE_SEQ)
# if defined(HAVE_CONST_TCP_QUEUE_SEQ)
#  define TCP_QUEUE_SEQ TCP_QUEUE_SEQ
# endif
#endif
#if !defined(TCP_REPAIR)
# if defined(HAVE_CONST_TCP_REPAIR)
#  define TCP_REPAIR TCP_REPAIR
# endif
#endif
#if !defined(TCP_REPAIR_OPTIONS)
# if defined(HAVE_CONST_TCP_REPAIR_OPTIONS)
#  define TCP_REPAIR_OPTIONS TCP_REPAIR_OPTIONS
# endif
#endif
#if !defined(TCP_REPAIR_QUEUE)
# if defined(HAVE_CONST_TCP_REPAIR_QUEUE)
#  define TCP_REPAIR_QUEUE TCP_REPAIR_QUEUE
# endif
#endif
#if !defined(TCP_THIN_DUPACK)
# if defined(HAVE_CONST_TCP_THIN_DUPACK)
#  define TCP_THIN_DUPACK TCP_THIN_DUPACK
# endif
#endif
#if !defined(TCP_THIN_LINEAR_TIMEOUTS)
# if defined(HAVE_CONST_TCP_THIN_LINEAR_TIMEOUTS)
#  define TCP_THIN_LINEAR_TIMEOUTS TCP_THIN_LINEAR_TIMEOUTS
# endif
#endif
#if !defined(TCP_TIMESTAMP)
# if defined(HAVE_CONST_TCP_TIMESTAMP)
#  define TCP_TIMESTAMP TCP_TIMESTAMP
# endif
#endif
#if !defined(TCP_USER_TIMEOUT)
# if defined(HAVE_CONST_TCP_USER_TIMEOUT)
#  define TCP_USER_TIMEOUT TCP_USER_TIMEOUT
# endif
#endif
#if !defined(UDP_CORK)
# if defined(HAVE_CONST_UDP_CORK)
#  define UDP_CORK UDP_CORK
# endif
#endif
#if !defined(EAI_ADDRFAMILY)
# if defined(HAVE_CONST_EAI_ADDRFAMILY)
#  define EAI_ADDRFAMILY EAI_ADDRFAMILY
# endif
#endif
#if !defined(EAI_AGAIN)
# if defined(HAVE_CONST_EAI_AGAIN)
#  define EAI_AGAIN EAI_AGAIN
# endif
#endif
#if !defined(EAI_BADFLAGS)
# if defined(HAVE_CONST_EAI_BADFLAGS)
#  define EAI_BADFLAGS EAI_BADFLAGS
# endif
#endif
#if !defined(EAI_FAIL)
# if defined(HAVE_CONST_EAI_FAIL)
#  define EAI_FAIL EAI_FAIL
# endif
#endif
#if !defined(EAI_FAMILY)
# if defined(HAVE_CONST_EAI_FAMILY)
#  define EAI_FAMILY EAI_FAMILY
# endif
#endif
#if !defined(EAI_MEMORY)
# if defined(HAVE_CONST_EAI_MEMORY)
#  define EAI_MEMORY EAI_MEMORY
# endif
#endif
#if !defined(EAI_NODATA)
# if defined(HAVE_CONST_EAI_NODATA)
#  define EAI_NODATA EAI_NODATA
# endif
#endif
#if !defined(EAI_NONAME)
# if defined(HAVE_CONST_EAI_NONAME)
#  define EAI_NONAME EAI_NONAME
# endif
#endif
#if !defined(EAI_OVERFLOW)
# if defined(HAVE_CONST_EAI_OVERFLOW)
#  define EAI_OVERFLOW EAI_OVERFLOW
# endif
#endif
#if !defined(EAI_SERVICE)
# if defined(HAVE_CONST_EAI_SERVICE)
#  define EAI_SERVICE EAI_SERVICE
# endif
#endif
#if !defined(EAI_SOCKTYPE)
# if defined(HAVE_CONST_EAI_SOCKTYPE)
#  define EAI_SOCKTYPE EAI_SOCKTYPE
# endif
#endif
#if !defined(EAI_SYSTEM)
# if defined(HAVE_CONST_EAI_SYSTEM)
#  define EAI_SYSTEM EAI_SYSTEM
# endif
#endif
#if !defined(EAI_BADHINTS)
# if defined(HAVE_CONST_EAI_BADHINTS)
#  define EAI_BADHINTS EAI_BADHINTS
# endif
#endif
#if !defined(EAI_PROTOCOL)
# if defined(HAVE_CONST_EAI_PROTOCOL)
#  define EAI_PROTOCOL EAI_PROTOCOL
# endif
#endif
#if !defined(EAI_MAX)
# if defined(HAVE_CONST_EAI_MAX)
#  define EAI_MAX EAI_MAX
# endif
#endif
#if !defined(AI_PASSIVE)
# if defined(HAVE_CONST_AI_PASSIVE)
#  define AI_PASSIVE AI_PASSIVE
# endif
#endif
#if !defined(AI_CANONNAME)
# if defined(HAVE_CONST_AI_CANONNAME)
#  define AI_CANONNAME AI_CANONNAME
# endif
#endif
#if !defined(AI_NUMERICHOST)
# if defined(HAVE_CONST_AI_NUMERICHOST)
#  define AI_NUMERICHOST AI_NUMERICHOST
# endif
#endif
#if !defined(AI_NUMERICSERV)
# if defined(HAVE_CONST_AI_NUMERICSERV)
#  define AI_NUMERICSERV AI_NUMERICSERV
# endif
#endif
#if !defined(AI_MASK)
# if defined(HAVE_CONST_AI_MASK)
#  define AI_MASK AI_MASK
# endif
#endif
#if !defined(AI_ALL)
# if defined(HAVE_CONST_AI_ALL)
#  define AI_ALL AI_ALL
# endif
#endif
#if !defined(AI_V4MAPPED_CFG)
# if defined(HAVE_CONST_AI_V4MAPPED_CFG)
#  define AI_V4MAPPED_CFG AI_V4MAPPED_CFG
# endif
#endif
#if !defined(AI_ADDRCONFIG)
# if defined(HAVE_CONST_AI_ADDRCONFIG)
#  define AI_ADDRCONFIG AI_ADDRCONFIG
# endif
#endif
#if !defined(AI_V4MAPPED)
# if defined(HAVE_CONST_AI_V4MAPPED)
#  define AI_V4MAPPED AI_V4MAPPED
# endif
#endif
#if !defined(AI_DEFAULT)
# if defined(HAVE_CONST_AI_DEFAULT)
#  define AI_DEFAULT AI_DEFAULT
# endif
#endif
#if !defined(NI_MAXHOST)
# if defined(HAVE_CONST_NI_MAXHOST)
#  define NI_MAXHOST NI_MAXHOST
# endif
#endif
#if !defined(NI_MAXSERV)
# if defined(HAVE_CONST_NI_MAXSERV)
#  define NI_MAXSERV NI_MAXSERV
# endif
#endif
#if !defined(NI_NOFQDN)
# if defined(HAVE_CONST_NI_NOFQDN)
#  define NI_NOFQDN NI_NOFQDN
# endif
#endif
#if !defined(NI_NUMERICHOST)
# if defined(HAVE_CONST_NI_NUMERICHOST)
#  define NI_NUMERICHOST NI_NUMERICHOST
# endif
#endif
#if !defined(NI_NAMEREQD)
# if defined(HAVE_CONST_NI_NAMEREQD)
#  define NI_NAMEREQD NI_NAMEREQD
# endif
#endif
#if !defined(NI_NUMERICSERV)
# if defined(HAVE_CONST_NI_NUMERICSERV)
#  define NI_NUMERICSERV NI_NUMERICSERV
# endif
#endif
#if !defined(NI_DGRAM)
# if defined(HAVE_CONST_NI_DGRAM)
#  define NI_DGRAM NI_DGRAM
# endif
#endif
#if !defined(SHUT_RD)
# if defined(HAVE_CONST_SHUT_RD)
#  define SHUT_RD SHUT_RD
# else
#  define SHUT_RD 0
# endif
#endif
#if !defined(SHUT_WR)
# if defined(HAVE_CONST_SHUT_WR)
#  define SHUT_WR SHUT_WR
# else
#  define SHUT_WR 1
# endif
#endif
#if !defined(SHUT_RDWR)
# if defined(HAVE_CONST_SHUT_RDWR)
#  define SHUT_RDWR SHUT_RDWR
# else
#  define SHUT_RDWR 2
# endif
#endif
#if !defined(IPV6_JOIN_GROUP)
# if defined(HAVE_CONST_IPV6_JOIN_GROUP)
#  define IPV6_JOIN_GROUP IPV6_JOIN_GROUP
# endif
#endif
#if !defined(IPV6_LEAVE_GROUP)
# if defined(HAVE_CONST_IPV6_LEAVE_GROUP)
#  define IPV6_LEAVE_GROUP IPV6_LEAVE_GROUP
# endif
#endif
#if !defined(IPV6_MULTICAST_HOPS)
# if defined(HAVE_CONST_IPV6_MULTICAST_HOPS)
#  define IPV6_MULTICAST_HOPS IPV6_MULTICAST_HOPS
# endif
#endif
#if !defined(IPV6_MULTICAST_IF)
# if defined(HAVE_CONST_IPV6_MULTICAST_IF)
#  define IPV6_MULTICAST_IF IPV6_MULTICAST_IF
# endif
#endif
#if !defined(IPV6_MULTICAST_LOOP)
# if defined(HAVE_CONST_IPV6_MULTICAST_LOOP)
#  define IPV6_MULTICAST_LOOP IPV6_MULTICAST_LOOP
# endif
#endif
#if !defined(IPV6_UNICAST_HOPS)
# if defined(HAVE_CONST_IPV6_UNICAST_HOPS)
#  define IPV6_UNICAST_HOPS IPV6_UNICAST_HOPS
# endif
#endif
#if !defined(IPV6_V6ONLY)
# if defined(HAVE_CONST_IPV6_V6ONLY)
#  define IPV6_V6ONLY IPV6_V6ONLY
# endif
#endif
#if !defined(IPV6_CHECKSUM)
# if defined(HAVE_CONST_IPV6_CHECKSUM)
#  define IPV6_CHECKSUM IPV6_CHECKSUM
# endif
#endif
#if !defined(IPV6_DONTFRAG)
# if defined(HAVE_CONST_IPV6_DONTFRAG)
#  define IPV6_DONTFRAG IPV6_DONTFRAG
# endif
#endif
#if !defined(IPV6_DSTOPTS)
# if defined(HAVE_CONST_IPV6_DSTOPTS)
#  define IPV6_DSTOPTS IPV6_DSTOPTS
# endif
#endif
#if !defined(IPV6_HOPLIMIT)
# if defined(HAVE_CONST_IPV6_HOPLIMIT)
#  define IPV6_HOPLIMIT IPV6_HOPLIMIT
# endif
#endif
#if !defined(IPV6_HOPOPTS)
# if defined(HAVE_CONST_IPV6_HOPOPTS)
#  define IPV6_HOPOPTS IPV6_HOPOPTS
# endif
#endif
#if !defined(IPV6_NEXTHOP)
# if defined(HAVE_CONST_IPV6_NEXTHOP)
#  define IPV6_NEXTHOP IPV6_NEXTHOP
# endif
#endif
#if !defined(IPV6_PATHMTU)
# if defined(HAVE_CONST_IPV6_PATHMTU)
#  define IPV6_PATHMTU IPV6_PATHMTU
# endif
#endif
#if !defined(IPV6_PKTINFO)
# if defined(HAVE_CONST_IPV6_PKTINFO)
#  define IPV6_PKTINFO IPV6_PKTINFO
# endif
#endif
#if !defined(IPV6_RECVDSTOPTS)
# if defined(HAVE_CONST_IPV6_RECVDSTOPTS)
#  define IPV6_RECVDSTOPTS IPV6_RECVDSTOPTS
# endif
#endif
#if !defined(IPV6_RECVHOPLIMIT)
# if defined(HAVE_CONST_IPV6_RECVHOPLIMIT)
#  define IPV6_RECVHOPLIMIT IPV6_RECVHOPLIMIT
# endif
#endif
#if !defined(IPV6_RECVHOPOPTS)
# if defined(HAVE_CONST_IPV6_RECVHOPOPTS)
#  define IPV6_RECVHOPOPTS IPV6_RECVHOPOPTS
# endif
#endif
#if !defined(IPV6_RECVPKTINFO)
# if defined(HAVE_CONST_IPV6_RECVPKTINFO)
#  define IPV6_RECVPKTINFO IPV6_RECVPKTINFO
# endif
#endif
#if !defined(IPV6_RECVRTHDR)
# if defined(HAVE_CONST_IPV6_RECVRTHDR)
#  define IPV6_RECVRTHDR IPV6_RECVRTHDR
# endif
#endif
#if !defined(IPV6_RECVTCLASS)
# if defined(HAVE_CONST_IPV6_RECVTCLASS)
#  define IPV6_RECVTCLASS IPV6_RECVTCLASS
# endif
#endif
#if !defined(IPV6_RTHDR)
# if defined(HAVE_CONST_IPV6_RTHDR)
#  define IPV6_RTHDR IPV6_RTHDR
# endif
#endif
#if !defined(IPV6_RTHDRDSTOPTS)
# if defined(HAVE_CONST_IPV6_RTHDRDSTOPTS)
#  define IPV6_RTHDRDSTOPTS IPV6_RTHDRDSTOPTS
# endif
#endif
#if !defined(IPV6_RTHDR_TYPE_0)
# if defined(HAVE_CONST_IPV6_RTHDR_TYPE_0)
#  define IPV6_RTHDR_TYPE_0 IPV6_RTHDR_TYPE_0
# endif
#endif
#if !defined(IPV6_RECVPATHMTU)
# if defined(HAVE_CONST_IPV6_RECVPATHMTU)
#  define IPV6_RECVPATHMTU IPV6_RECVPATHMTU
# endif
#endif
#if !defined(IPV6_TCLASS)
# if defined(HAVE_CONST_IPV6_TCLASS)
#  define IPV6_TCLASS IPV6_TCLASS
# endif
#endif
#if !defined(IPV6_USE_MIN_MTU)
# if defined(HAVE_CONST_IPV6_USE_MIN_MTU)
#  define IPV6_USE_MIN_MTU IPV6_USE_MIN_MTU
# endif
#endif
#if !defined(INET_ADDRSTRLEN)
# if defined(HAVE_CONST_INET_ADDRSTRLEN)
#  define INET_ADDRSTRLEN INET_ADDRSTRLEN
# else
#  define INET_ADDRSTRLEN 16
# endif
#endif
#if !defined(INET6_ADDRSTRLEN)
# if defined(HAVE_CONST_INET6_ADDRSTRLEN)
#  define INET6_ADDRSTRLEN INET6_ADDRSTRLEN
# else
#  define INET6_ADDRSTRLEN 46
# endif
#endif
#if !defined(IFNAMSIZ)
# if defined(HAVE_CONST_IFNAMSIZ)
#  define IFNAMSIZ IFNAMSIZ
# endif
#endif
#if !defined(IF_NAMESIZE)
# if defined(HAVE_CONST_IF_NAMESIZE)
#  define IF_NAMESIZE IF_NAMESIZE
# endif
#endif
#if !defined(SOMAXCONN)
# if defined(HAVE_CONST_SOMAXCONN)
#  define SOMAXCONN SOMAXCONN
# else
#  define SOMAXCONN 5
# endif
#endif
#if !defined(SCM_RIGHTS)
# if defined(HAVE_CONST_SCM_RIGHTS)
#  define SCM_RIGHTS SCM_RIGHTS
# endif
#endif
#if !defined(SCM_TIMESTAMP)
# if defined(HAVE_CONST_SCM_TIMESTAMP)
#  define SCM_TIMESTAMP SCM_TIMESTAMP
# endif
#endif
#if !defined(SCM_TIMESTAMPNS)
# if defined(HAVE_CONST_SCM_TIMESTAMPNS)
#  define SCM_TIMESTAMPNS SCM_TIMESTAMPNS
# endif
#endif
#if !defined(SCM_TIMESTAMPING)
# if defined(HAVE_CONST_SCM_TIMESTAMPING)
#  define SCM_TIMESTAMPING SCM_TIMESTAMPING
# endif
#endif
#if !defined(SCM_BINTIME)
# if defined(HAVE_CONST_SCM_BINTIME)
#  define SCM_BINTIME SCM_BINTIME
# endif
#endif
#if !defined(SCM_CREDENTIALS)
# if defined(HAVE_CONST_SCM_CREDENTIALS)
#  define SCM_CREDENTIALS SCM_CREDENTIALS
# endif
#endif
#if !defined(SCM_CREDS)
# if defined(HAVE_CONST_SCM_CREDS)
#  define SCM_CREDS SCM_CREDS
# endif
#endif
#if !defined(SCM_UCRED)
# if defined(HAVE_CONST_SCM_UCRED)
#  define SCM_UCRED SCM_UCRED
# endif
#endif
#if !defined(SCM_WIFI_STATUS)
# if defined(HAVE_CONST_SCM_WIFI_STATUS)
#  define SCM_WIFI_STATUS SCM_WIFI_STATUS
# endif
#endif
#if !defined(LOCAL_PEERCRED)
# if defined(HAVE_CONST_LOCAL_PEERCRED)
#  define LOCAL_PEERCRED LOCAL_PEERCRED
# endif
#endif
#if !defined(LOCAL_CREDS)
# if defined(HAVE_CONST_LOCAL_CREDS)
#  define LOCAL_CREDS LOCAL_CREDS
# endif
#endif
#if !defined(LOCAL_CONNWAIT)
# if defined(HAVE_CONST_LOCAL_CONNWAIT)
#  define LOCAL_CONNWAIT LOCAL_CONNWAIT
# endif
#endif
#if !defined(IFF_802_1Q_VLAN)
# if defined(HAVE_CONST_IFF_802_1Q_VLAN)
#  define IFF_802_1Q_VLAN IFF_802_1Q_VLAN
# endif
#endif
#if !defined(IFF_ALLMULTI)
# if defined(HAVE_CONST_IFF_ALLMULTI)
#  define IFF_ALLMULTI IFF_ALLMULTI
# endif
#endif
#if !defined(IFF_ALTPHYS)
# if defined(HAVE_CONST_IFF_ALTPHYS)
#  define IFF_ALTPHYS IFF_ALTPHYS
# endif
#endif
#if !defined(IFF_AUTOMEDIA)
# if defined(HAVE_CONST_IFF_AUTOMEDIA)
#  define IFF_AUTOMEDIA IFF_AUTOMEDIA
# endif
#endif
#if !defined(IFF_BONDING)
# if defined(HAVE_CONST_IFF_BONDING)
#  define IFF_BONDING IFF_BONDING
# endif
#endif
#if !defined(IFF_BRIDGE_PORT)
# if defined(HAVE_CONST_IFF_BRIDGE_PORT)
#  define IFF_BRIDGE_PORT IFF_BRIDGE_PORT
# endif
#endif
#if !defined(IFF_BROADCAST)
# if defined(HAVE_CONST_IFF_BROADCAST)
#  define IFF_BROADCAST IFF_BROADCAST
# endif
#endif
#if !defined(IFF_CANTCONFIG)
# if defined(HAVE_CONST_IFF_CANTCONFIG)
#  define IFF_CANTCONFIG IFF_CANTCONFIG
# endif
#endif
#if !defined(IFF_DEBUG)
# if defined(HAVE_CONST_IFF_DEBUG)
#  define IFF_DEBUG IFF_DEBUG
# endif
#endif
#if !defined(IFF_DISABLE_NETPOLL)
# if defined(HAVE_CONST_IFF_DISABLE_NETPOLL)
#  define IFF_DISABLE_NETPOLL IFF_DISABLE_NETPOLL
# endif
#endif
#if !defined(IFF_DONT_BRIDGE)
# if defined(HAVE_CONST_IFF_DONT_BRIDGE)
#  define IFF_DONT_BRIDGE IFF_DONT_BRIDGE
# endif
#endif
#if !defined(IFF_DORMANT)
# if defined(HAVE_CONST_IFF_DORMANT)
#  define IFF_DORMANT IFF_DORMANT
# endif
#endif
#if !defined(IFF_DRV_OACTIVE)
# if defined(HAVE_CONST_IFF_DRV_OACTIVE)
#  define IFF_DRV_OACTIVE IFF_DRV_OACTIVE
# endif
#endif
#if !defined(IFF_DRV_RUNNING)
# if defined(HAVE_CONST_IFF_DRV_RUNNING)
#  define IFF_DRV_RUNNING IFF_DRV_RUNNING
# endif
#endif
#if !defined(IFF_DYING)
# if defined(HAVE_CONST_IFF_DYING)
#  define IFF_DYING IFF_DYING
# endif
#endif
#if !defined(IFF_DYNAMIC)
# if defined(HAVE_CONST_IFF_DYNAMIC)
#  define IFF_DYNAMIC IFF_DYNAMIC
# endif
#endif
#if !defined(IFF_EBRIDGE)
# if defined(HAVE_CONST_IFF_EBRIDGE)
#  define IFF_EBRIDGE IFF_EBRIDGE
# endif
#endif
#if !defined(IFF_ECHO)
# if defined(HAVE_CONST_IFF_ECHO)
#  define IFF_ECHO IFF_ECHO
# endif
#endif
#if !defined(IFF_ISATAP)
# if defined(HAVE_CONST_IFF_ISATAP)
#  define IFF_ISATAP IFF_ISATAP
# endif
#endif
#if !defined(IFF_LINK0)
# if defined(HAVE_CONST_IFF_LINK0)
#  define IFF_LINK0 IFF_LINK0
# endif
#endif
#if !defined(IFF_LINK1)
# if defined(HAVE_CONST_IFF_LINK1)
#  define IFF_LINK1 IFF_LINK1
# endif
#endif
#if !defined(IFF_LINK2)
# if defined(HAVE_CONST_IFF_LINK2)
#  define IFF_LINK2 IFF_LINK2
# endif
#endif
#if !defined(IFF_LIVE_ADDR_CHANGE)
# if defined(HAVE_CONST_IFF_LIVE_ADDR_CHANGE)
#  define IFF_LIVE_ADDR_CHANGE IFF_LIVE_ADDR_CHANGE
# endif
#endif
#if !defined(IFF_LOOPBACK)
# if defined(HAVE_CONST_IFF_LOOPBACK)
#  define IFF_LOOPBACK IFF_LOOPBACK
# endif
#endif
#if !defined(IFF_LOWER_UP)
# if defined(HAVE_CONST_IFF_LOWER_UP)
#  define IFF_LOWER_UP IFF_LOWER_UP
# endif
#endif
#if !defined(IFF_MACVLAN_PORT)
# if defined(HAVE_CONST_IFF_MACVLAN_PORT)
#  define IFF_MACVLAN_PORT IFF_MACVLAN_PORT
# endif
#endif
#if !defined(IFF_MASTER)
# if defined(HAVE_CONST_IFF_MASTER)
#  define IFF_MASTER IFF_MASTER
# endif
#endif
#if !defined(IFF_MASTER_8023AD)
# if defined(HAVE_CONST_IFF_MASTER_8023AD)
#  define IFF_MASTER_8023AD IFF_MASTER_8023AD
# endif
#endif
#if !defined(IFF_MASTER_ALB)
# if defined(HAVE_CONST_IFF_MASTER_ALB)
#  define IFF_MASTER_ALB IFF_MASTER_ALB
# endif
#endif
#if !defined(IFF_MASTER_ARPMON)
# if defined(HAVE_CONST_IFF_MASTER_ARPMON)
#  define IFF_MASTER_ARPMON IFF_MASTER_ARPMON
# endif
#endif
#if !defined(IFF_MONITOR)
# if defined(HAVE_CONST_IFF_MONITOR)
#  define IFF_MONITOR IFF_MONITOR
# endif
#endif
#if !defined(IFF_MULTICAST)
# if defined(HAVE_CONST_IFF_MULTICAST)
#  define IFF_MULTICAST IFF_MULTICAST
# endif
#endif
#if !defined(IFF_NOARP)
# if defined(HAVE_CONST_IFF_NOARP)
#  define IFF_NOARP IFF_NOARP
# endif
#endif
#if !defined(IFF_NOTRAILERS)
# if defined(HAVE_CONST_IFF_NOTRAILERS)
#  define IFF_NOTRAILERS IFF_NOTRAILERS
# endif
#endif
#if !defined(IFF_OACTIVE)
# if defined(HAVE_CONST_IFF_OACTIVE)
#  define IFF_OACTIVE IFF_OACTIVE
# endif
#endif
#if !defined(IFF_OVS_DATAPATH)
# if defined(HAVE_CONST_IFF_OVS_DATAPATH)
#  define IFF_OVS_DATAPATH IFF_OVS_DATAPATH
# endif
#endif
#if !defined(IFF_POINTOPOINT)
# if defined(HAVE_CONST_IFF_POINTOPOINT)
#  define IFF_POINTOPOINT IFF_POINTOPOINT
# endif
#endif
#if !defined(IFF_PORTSEL)
# if defined(HAVE_CONST_IFF_PORTSEL)
#  define IFF_PORTSEL IFF_PORTSEL
# endif
#endif
#if !defined(IFF_PPROMISC)
# if defined(HAVE_CONST_IFF_PPROMISC)
#  define IFF_PPROMISC IFF_PPROMISC
# endif
#endif
#if !defined(IFF_PROMISC)
# if defined(HAVE_CONST_IFF_PROMISC)
#  define IFF_PROMISC IFF_PROMISC
# endif
#endif
#if !defined(IFF_RENAMING)
# if defined(HAVE_CONST_IFF_RENAMING)
#  define IFF_RENAMING IFF_RENAMING
# endif
#endif
#if !defined(IFF_ROUTE)
# if defined(HAVE_CONST_IFF_ROUTE)
#  define IFF_ROUTE IFF_ROUTE
# endif
#endif
#if !defined(IFF_RUNNING)
# if defined(HAVE_CONST_IFF_RUNNING)
#  define IFF_RUNNING IFF_RUNNING
# endif
#endif
#if !defined(IFF_SIMPLEX)
# if defined(HAVE_CONST_IFF_SIMPLEX)
#  define IFF_SIMPLEX IFF_SIMPLEX
# endif
#endif
#if !defined(IFF_SLAVE)
# if defined(HAVE_CONST_IFF_SLAVE)
#  define IFF_SLAVE IFF_SLAVE
# endif
#endif
#if !defined(IFF_SLAVE_INACTIVE)
# if defined(HAVE_CONST_IFF_SLAVE_INACTIVE)
#  define IFF_SLAVE_INACTIVE IFF_SLAVE_INACTIVE
# endif
#endif
#if !defined(IFF_SLAVE_NEEDARP)
# if defined(HAVE_CONST_IFF_SLAVE_NEEDARP)
#  define IFF_SLAVE_NEEDARP IFF_SLAVE_NEEDARP
# endif
#endif
#if !defined(IFF_SMART)
# if defined(HAVE_CONST_IFF_SMART)
#  define IFF_SMART IFF_SMART
# endif
#endif
#if !defined(IFF_STATICARP)
# if defined(HAVE_CONST_IFF_STATICARP)
#  define IFF_STATICARP IFF_STATICARP
# endif
#endif
#if !defined(IFF_SUPP_NOFCS)
# if defined(HAVE_CONST_IFF_SUPP_NOFCS)
#  define IFF_SUPP_NOFCS IFF_SUPP_NOFCS
# endif
#endif
#if !defined(IFF_TEAM_PORT)
# if defined(HAVE_CONST_IFF_TEAM_PORT)
#  define IFF_TEAM_PORT IFF_TEAM_PORT
# endif
#endif
#if !defined(IFF_TX_SKB_SHARING)
# if defined(HAVE_CONST_IFF_TX_SKB_SHARING)
#  define IFF_TX_SKB_SHARING IFF_TX_SKB_SHARING
# endif
#endif
#if !defined(IFF_UNICAST_FLT)
# if defined(HAVE_CONST_IFF_UNICAST_FLT)
#  define IFF_UNICAST_FLT IFF_UNICAST_FLT
# endif
#endif
#if !defined(IFF_UP)
# if defined(HAVE_CONST_IFF_UP)
#  define IFF_UP IFF_UP
# endif
#endif
#if !defined(IFF_WAN_HDLC)
# if defined(HAVE_CONST_IFF_WAN_HDLC)
#  define IFF_WAN_HDLC IFF_WAN_HDLC
# endif
#endif
#if !defined(IFF_XMIT_DST_RELEASE)
# if defined(HAVE_CONST_IFF_XMIT_DST_RELEASE)
#  define IFF_XMIT_DST_RELEASE IFF_XMIT_DST_RELEASE
# endif
#endif
#if !defined(IFF_VOLATILE)
# if defined(HAVE_CONST_IFF_VOLATILE)
#  define IFF_VOLATILE IFF_VOLATILE
# endif
#endif
#if !defined(IFF_CANTCHANGE)
# if defined(HAVE_CONST_IFF_CANTCHANGE)
#  define IFF_CANTCHANGE IFF_CANTCHANGE
# endif
#endif

int rsock_family_to_int(const char *str, long len, int *valp);

int rsock_socktype_to_int(const char *str, long len, int *valp);

int rsock_ipproto_to_int(const char *str, long len, int *valp);

int rsock_unknown_level_to_int(const char *str, long len, int *valp);

int rsock_ip_level_to_int(const char *str, long len, int *valp);

int rsock_so_optname_to_int(const char *str, long len, int *valp);

int rsock_ip_optname_to_int(const char *str, long len, int *valp);

#ifdef IPPROTO_IPV6
int rsock_ipv6_optname_to_int(const char *str, long len, int *valp);
#endif

int rsock_tcp_optname_to_int(const char *str, long len, int *valp);

int rsock_udp_optname_to_int(const char *str, long len, int *valp);

int rsock_shutdown_how_to_int(const char *str, long len, int *valp);

int rsock_scm_optname_to_int(const char *str, long len, int *valp);

ID rsock_intern_family(int val);

ID rsock_intern_family_noprefix(int val);

ID rsock_intern_protocol_family(int val);

ID rsock_intern_socktype(int val);

ID rsock_intern_ipproto(int val);

ID rsock_intern_iplevel(int val);

ID rsock_intern_so_optname(int val);

ID rsock_intern_ip_optname(int val);

ID rsock_intern_ipv6_optname(int val);

ID rsock_intern_tcp_optname(int val);

ID rsock_intern_udp_optname(int val);

ID rsock_intern_scm_optname(int val);

ID rsock_intern_local_optname(int val);

